package org.netbeans.modules.cnd.dwarfdump.section;
import java.io.IOException;
import java.util.ArrayList;

import org.netbeans.modules.cnd.dwarfdump.dwarfconsts.LNE;
import org.netbeans.modules.cnd.dwarfdump.dwarfconsts.LNS;
import org.netbeans.modules.cnd.dwarfdump.reader.ElfReader;

public class DebugLineStateMachine {
	// DWARF3 State Machine Registers (6.2.2)
	private DebugLineRow regs;
	// Relevant information from line number program header
	private int minimum_instruction_length;
	private boolean default_is_stmt;
	private int line_base;
	private int line_range;
	private int opcode_base;
	private int[] standard_opcode_lengths;
	private ArrayList<DebugLineFileEntry> fileEntries;
	// Offset of the last byte in the line number program
	private long endOffset;
	private ElfReader reader;
	
	@SuppressWarnings("unchecked")
	public DebugLineStateMachine(int minimum_instruction_length,
			boolean default_is_stmt,
			int line_base,
			int line_range,
			int opcode_base,
			int[] standard_opcode_lengths,
			ArrayList<DebugLineFileEntry> fileEntries,
			long length,
			ElfReader reader) throws IOException
	{
		this.minimum_instruction_length = minimum_instruction_length;
		this.default_is_stmt = default_is_stmt;
		this.line_base = line_base;
		this.line_range = line_range;
		this.opcode_base = opcode_base;
		this.standard_opcode_lengths = standard_opcode_lengths.clone();
		this.fileEntries = (ArrayList<DebugLineFileEntry>)fileEntries.clone();
		this.endOffset = reader.getFilePointer() + length;
		this.reader = reader;
		// TODO check standard_opcode_lengths agree with those we know about
		regs = new DebugLineRow();
		reset();
	}
	
	public boolean hasNext() throws IOException
	{
		return reader.getFilePointer() != endOffset;
	}
	
	/**
	 * Step to the next row of the matrix
	 * @param reader
	 */
	public void step() throws IOException
	{
		if (!hasNext()) {
			throw new IllegalStateException("Trying to step past the end of the line number program");
		}
		if (regs.end_sequence) {
			// All registers are reset following the end of a sequence 
			reset();
		} else {
			// These registers should be reset after each row
			regs.basic_block = false;
			regs.prologue_end = false;
			regs.epilogue_begin = false;
		}
		// Read until new row is reached
		while (!readInstruction(reader));
	}
	
	/**
	 * Reset all registers to their default state
	 */
	private void reset()
	{
		regs.address = 0;
		regs.file = 1;
		regs.line = 1;
		regs.column = 0;
		regs.is_stmt = default_is_stmt;
		regs.basic_block = false;
		regs.end_sequence = false;
		regs.prologue_end = false;
		regs.epilogue_begin = false;
		regs.isa = 0;
	}
	
	/**
	 * Reads and executes a single state machine instruction
	 * @param reader The DwarfReader
	 * @return Whether a new row has been reached
	 */
	private boolean readInstruction(ElfReader reader) throws IOException
	{
		boolean newRow = false;
		int opcode = reader.readUnsignedByte();
		if (opcode == 0) {
			// Extended opcode, see DWARF3 6.2.5.3
			newRow = readExtendedInstruction(reader);
		} else if (opcode < opcode_base) {
			// Standard opcode, see DWARF3 6.2.5.2
			newRow = readStandardInstruction(reader, opcode);
		} else {
			// Special opcode, see DWARF3 6.2.5.1
			int adjustedOpcode = opcode - opcode_base;
			int addressIncrement = (adjustedOpcode / line_range) * minimum_instruction_length;
			int lineIncrement = line_base + (adjustedOpcode % line_range);
			regs.line += lineIncrement;
			regs.address += addressIncrement;
			newRow = true;
		}
		return newRow;
	}
	
	/**
	 * Reads the arguments for a single standard state machine instruction
	 * and executions the instruction. 
	 * @param reader
	 * @param opcode The opcode for the instruction
	 * @return Whether a new row has been reached
	 * @throws IOException
	 */
	private boolean readStandardInstruction(ElfReader reader, int opcode) throws IOException {
		boolean newRow = false;
		LNS op = LNS.get(opcode);
		if (op != null) {
			switch (op) {
			case DW_LNS_copy:
				newRow = true;
				break;
			case DW_LNS_advance_pc:
				regs.address += (long)reader.readUnsignedLEB128() * minimum_instruction_length;
				break;
			case DW_LNS_advance_line:
				regs.line += reader.readSignedLEB128();
				break;
			case DW_LNS_set_file:
				regs.file = reader.readUnsignedLEB128();
				break;
			case DW_LNS_set_column:
				regs.column = reader.readUnsignedLEB128();
				break;
			case DW_LNS_negate_stmt:
				regs.is_stmt = !regs.is_stmt;
				break;
			case DW_LNS_set_basic_block:
				regs.basic_block = true;
				break;
			case DW_LNS_const_add_pc:
				// Use the address increment value corresponding to special
				// opcode 255
				int adjustedOpcode = 255 - opcode_base;
				int addressIncrement = (adjustedOpcode / line_range) * minimum_instruction_length;
				regs.address += addressIncrement;
				break;
			case DW_LNS_fixed_advance_pc:
				regs.address += reader.readUnsignedShort();
				break;
			case DW_LNS_set_prologue_end:
				regs.prologue_end = true;
				break;
			case DW_LNS_set_epilogue_begin:
				regs.epilogue_begin = true;
				break;
			case DW_LNS_set_isa:
				regs.isa = reader.readUnsignedLEB128();
				break;
			}
		} else {
			// Skip any unrecognised standard instructions
			int numArgs = standard_opcode_lengths[opcode - 1];
			for (int i = 0; i < numArgs; i++) {
				reader.readSignedLEB128();
			}
		}
		return newRow;
	}
	
	private boolean readExtendedInstruction(ElfReader reader) throws IOException {
		int numBytes = reader.readUnsignedLEB128();
		int opcode = reader.readUnsignedByte();
		boolean newRow = false;
		LNE op = LNE.get(opcode);
		if (op != null) {
			// TODO check numBytes is correct
			switch (op) {
			case DW_LNE_end_sequence:
				regs.end_sequence = true;
				newRow = true;
				break;
			case DW_LNE_set_address:
				regs.address = reader.readAddress();
				break;
			case DW_LNE_define_file:
				fileEntries.add(new DebugLineFileEntry(reader.readString(),
						reader.readUnsignedLEB128(),
						reader.readUnsignedLEB128(),
						reader.readUnsignedLEB128()));
				break;
			}
		} else {
			// Skip any unrecognised extended instructions
			reader.skipBytes(numBytes);
		}
		return newRow;
	}
	
	/**
	 * Getters
	 */
	
	public long getAddress() {
		return regs.address;
	}
	
	public DebugLineFileEntry getFile() {
		return fileEntries.get((int)regs.file);
	}
	
	public long getLine() {
		return regs.line;
	}
	
	public long getColumn() {
		return regs.column;
	}
	
	public boolean isStmt() {
		return regs.is_stmt;
	}
	
	public boolean isBasicBlock() {
		return regs.basic_block;
	}
	
	public boolean isEndSequence() {
		return regs.end_sequence;
	}
	
	public boolean isPrologueEnd() {
		return regs.prologue_end;
	}
	
	public boolean isEpilogueBegin() {
		return regs.epilogue_begin;
	}
	
	public long getIsa() {
		return regs.isa;
	}
	
	/**
	 * Returns a copy of the current row
	 * @return A copy of the current row
	 */
	public DebugLineRow getRow() {
		return (DebugLineRow) regs.clone();
	}
	
	@SuppressWarnings("unchecked")
	public ArrayList<DebugLineFileEntry> getFileEntries() {
		return (ArrayList<DebugLineFileEntry>) fileEntries.clone();
	}
}
